@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Red Hat, Inc.
+ * Copyright (C) 2014-2016 Red Hat, Inc.
  *
  * This file is part of ocserv.
  *
@@ -115,7 +115,7 @@ static void append_stats(rc_handle *rh, VALUE_PAIR **send, stats_st *stats)
 	return;
 }
 
-static void append_acct_standard(rc_handle *rh, const common_auth_info_st *ai, VALUE_PAIR **send)
+static void append_acct_standard(rc_handle *rh, const common_acct_info_st *ai, VALUE_PAIR **send)
 {
 	uint32_t i;
 
@@ -185,16 +185,10 @@ static void append_acct_standard(rc_handle *rh, const common_auth_info_st *ai, V
 		return;
 	}
 
-	if (ai->user_agent[0] != 0) {
-	    if (rc_avpair_add(rh, send, PW_CONNECT_INFO, ai->user_agent, -1, 0) == NULL) {
-		    return;
-	    }
-	}
-
 	return;
 }
 
-static void radius_acct_session_stats(unsigned auth_method, const common_auth_info_st *ai, stats_st *stats)
+static void radius_acct_session_stats(unsigned auth_method, const common_acct_info_st *ai, stats_st *stats)
 {
 int ret;
 uint32_t status_type;
@@ -227,7 +221,7 @@ VALUE_PAIR *send = NULL, *recvd = NULL;
 	return;
 }
 
-static int radius_acct_open_session(unsigned auth_method, const common_auth_info_st *ai, const void *sid, unsigned sid_size)
+static int radius_acct_open_session(unsigned auth_method, const common_acct_info_st *ai, const void *sid, unsigned sid_size)
 {
 int ret;
 uint32_t status_type;
@@ -247,6 +241,13 @@ VALUE_PAIR *send = NULL, *recvd = NULL;
 		goto cleanup;
 	}
 
+	if (ai->user_agent[0] != 0) {
+		if (rc_avpair_add(rh, &send, PW_CONNECT_INFO, ai->user_agent, -1, 0) == NULL) {
+			ret = -1;
+			goto cleanup;
+		}
+	}
+
 	append_acct_standard(rh, ai, &send);
 
 	ret = rc_aaa(rh, ai->id, send, &recvd, NULL, 1, PW_ACCOUNTING_REQUEST);
@@ -266,7 +267,7 @@ VALUE_PAIR *send = NULL, *recvd = NULL;
 	return ret;
 }
 
-static void radius_acct_close_session(unsigned auth_method, const common_auth_info_st *ai, stats_st *stats, unsigned discon_reason)
+static void radius_acct_close_session(unsigned auth_method, const common_acct_info_st *ai, stats_st *stats, unsigned discon_reason)
 {
 int ret;
 uint32_t status_type;
