@@ -22,11 +22,11 @@
 #define RADIUS_H
 
 #include <sec-mod-auth.h>
-#include <base64.h>
 
 struct radius_ctx_st {
 	char username[MAX_USERNAME_SIZE*2];
 	char groupname[MAX_GROUPNAME_SIZE];
+	char user_agent[MAX_AGENT_NAME];
 
 	char remote_ip[MAX_IP_STR];
 	char our_ip[MAX_IP_STR];
