@@ -136,7 +136,7 @@ if test ! -f $DOCKER_DIR/Dockerfile;then
 fi
 
 rm -f $DOCKER_DIR/ocserv $DOCKER_DIR/ocpasswd $DOCKER_DIR/occtl $DOCKER_DIR/ocserv-fw
-cp ../src/ocserv ../src/ocserv-fw ../src/ocpasswd ../src/occtl $DOCKER_DIR/
+cp ../src/ocserv ../src/ocserv-fw ../src/ocpasswd/ocpasswd ../src/occtl/occtl $DOCKER_DIR/
 
 echo "Creating image $IMAGE"
 $DOCKER build -t $IMAGE $DOCKER_DIR/
